from typing import Any, List, Optional, Set, Union

import asyncio
import logging
from contextlib import suppress
from urllib.parse import urlparse

from aiokafka import AIOKafkaConsumer, AIOKafkaProducer

from rhubarb.backends.base import BaseBackend
from rhubarb.backends.exceptions import UnsubscribeError
from rhubarb.event import Event


class KafkaBackend(BaseBackend):
    def __init__(self, url: str):
        """aiokafka backend

        :param url: the kafka URL to connect to
        :type url: str
        """
        self._servers: list[str] = [urlparse(url).netloc]
        self._channels: set[str] = set()
        self._lock = asyncio.Lock()
        self._listen_queue: asyncio.Queue[Union[Event, None]] = asyncio.Queue()
        self._consumer_reader_task = None
        self.logger = logging.getLogger("RedisBackend")

    async def connect(self) -> None:
        """Connects the producer to kafka backend"""
        self.logger.info("Connecting to %s", self._servers)
        self._producer = AIOKafkaProducer(bootstrap_servers=self._servers)
        await self._producer.start()
        self.logger.info("Connected to %s", self._servers)

    async def disconnect(self) -> None:
        """Disconnects and cleans up consumers"""
        self.logger.info("Disconnecting from %s", self._servers)
        await self._producer.stop()
        await self._stop_consumer()
        self.logger.info("Disconnected from %s", self._servers)

    async def _start_consumer(self):
        """Internal start consumer method creating a kafka consumer for the requested channels"""
        self._consumer = AIOKafkaConsumer(
            *self._channels, bootstrap_servers=self._servers
        )
        await self._consumer.start()
        self._consumer_reader_task: asyncio.Task[None] = asyncio.create_task(  # type: ignore
            self._reader()
        )

    async def _stop_consumer(self):
        """Stops the consumer task and the AIOKafkaConsumer object"""
        if self._consumer_reader_task:
            self._consumer_reader_task.cancel()  # type: ignore
            with suppress(asyncio.exceptions.CancelledError):
                await self._consumer_reader_task
            await self._consumer.stop()

    async def subscribe(self, channel: str) -> None:
        """Adds the provided channel to the set of ``self.channels`` and restarts the consumer

        :param channel:
        :type channel: str
        """
        async with self._lock:
            await self._stop_consumer()
            self._channels.add(channel)
            await self._start_consumer()

    async def unsubscribe(self, channel: str) -> None:
        """Removes the channel from the ``channels`` and re-creates the consumer

        :param channel: the channel to unsubscribe
        :type channel: str
        """
        self.logger.info("Unsubscribing from %s", channel)
        async with self._lock:
            if channel in self._channels:
                await self._stop_consumer()
                self._channels.remove(channel)
                if self._channels:
                    await self._start_consumer()
            else:
                self.logger.warning("Unknown channel %s", channel)
                raise UnsubscribeError(f"Unknown channel {channel}")

    async def publish(self, channel: str, message: Any) -> None:
        """Using the created ``self._producer`` publish a message to the provided channel

        :param channel: Channel name to use to publish the message to
        :type channel: str
        :param message: Data that is to be published to the channel
        :type message: Any
        """
        self.logger.debug("Publishing message %s to channel %s", message, channel)
        await self._producer.send_and_wait(channel, message.encode("utf8"))

    async def _reader(self) -> None:
        """Read data from the consumer and put events into a queue to be read by the caller."""
        async for message in self._consumer:
            self.logger.debug(
                "Read message %s from channel: %s",
                message.value.decode("utf8"),
                message.topic,
            )
            event = Event(channel=message.topic, message=message.value.decode("utf8"))
            self._listen_queue.put_nowait(event)

    async def next_event(self) -> Optional[Event]:
        """Return the next event from the queue that was read from all channels"""
        self.logger.debug("Getting next event from kafka...")
        return await self._listen_queue.get()



from typing import Any

from abc import ABC, abstractmethod

from rhubarb.event import Event


class BaseBackend(ABC):
    @abstractmethod
    def __init__(self, url: str) -> None:
        """
        :param url: URL of the backend service
        :type url: str
        """

    @abstractmethod
    async def connect(self) -> None:
        """Connect to queue service"""

    @abstractmethod
    async def disconnect(self) -> None:
        """Disconnect from queue service"""

    @abstractmethod
    async def subscribe(self, channel: str) -> None:
        """Subscribe to a channel

        :param channel: name of the channel in the queue to subscribe to
        :type channel: str
        """

    @abstractmethod
    async def unsubscribe(self, channel: str) -> None:
        """Unsubscribe from a channel

        :param channel: name of the channel in the queue to subscribe to
        :type channel: str
        """

    @abstractmethod
    async def publish(self, channel: str, message: Any) -> None:
        """Publish a message to a given channel

        :param channel: name of the channel in the queue to publish the message to
        :type channel: str
        :param message: the message to publish to the channel
        :type channel: str
        """

    @abstractmethod
    async def next_event(self) -> Event:
        """Get the next published Event from the queue

        :return:
        :rtype: Event
        """



from typing import Any, Dict, List, Union

import asyncio
import logging
from asyncio import exceptions
from collections import namedtuple

import aioredis
import async_timeout

from rhubarb.backends.base import BaseBackend
from rhubarb.backends.exceptions import UnsubscribeError
from rhubarb.event import Event

STOPWORD = "STOP"

Subscription = namedtuple("Subscription", ["pubsub", "task"])


class RedisBackend(BaseBackend):
    def __init__(self, url: str):
        """Backend for the message queue using ``aioredis`` to interface with redis

        :param url: The URL of the redis instance
        :type url: str
        """
        self.url: str = url
        self._channels: dict[str, Subscription] = {}
        self._listen_queue: asyncio.Queue[Union[Event, None]] = asyncio.Queue()
        self.logger = logging.getLogger("RedisBackend")

    async def connect(self) -> None:
        """Connection to the configured Redis URL using ``aioredis``.
        Execute a `ping` to check if the connection is valid.
        """
        self.logger.info("Connecting to %s", self.url)
        self._redis = aioredis.from_url(
            self.url, encoding="utf-8", decode_responses=True
        )
        await self._redis.ping()  # causes aioredis to connect immediately
        self.logger.info("Connected to %s", self.url)

    async def disconnect(self) -> None:
        """Gracefully close the queue and end the tasks, the redis connection gracefully closes itself during garbage collection."""
        self.logger.info("Disconnecting from %s", self.url)
        for channel, subscription in self._channels.items():
            await self.publish(channel, STOPWORD)
            await subscription.pubsub.unsubscribe(channel)
            subscription.task.cancel()
            await subscription.task
        self._listen_queue.put_nowait(None)
        self.logger.info("Disconnected from %s", self.url)

    async def _reader(self, channel: aioredis.client.PubSub) -> None:
        """Read data from the passed channel object and put events into a queue to be read by the caller.

        :param channel: The channel subscribed to
        :type channel: aioredis.client.PubSub
        """
        while True:
            try:
                async with async_timeout.timeout(1):
                    if (
                        message := await channel.get_message(
                            ignore_subscribe_messages=True, timeout=1.0
                        )
                    ) is not None:
                        self.logger.debug(
                            "Read message %s from channel: %s",
                            message["channel"],
                            message["data"],
                        )
                        if message["data"] == STOPWORD:
                            self.logger.info("Reader received stop message")
                            break
                        event = Event(
                            channel=message["channel"], message=message["data"]
                        )
                        self._listen_queue.put_nowait(event)
            except asyncio.CancelledError:
                break
            except asyncio.TimeoutError:
                pass

    async def subscribe(self, channel: str) -> None:
        """Subscribe to the passed channel

        :param channel: channel name to subscribe to
        :type channel: str
        """
        self.logger.info("Subscribing to %s", channel)
        if channel in self._channels:
            self.logger.warning("Already subscribed to %s", channel)
        else:
            _pubsub = self._redis.pubsub()
            await _pubsub.subscribe(channel)
            task = asyncio.create_task(self._reader(_pubsub))
            self._channels[channel] = Subscription(_pubsub, task)

    async def unsubscribe(self, channel: str) -> None:
        """Unsubscribe from the passed channel

        :param channel: channel name to subscribe to
        :type channel: str
        :raises UnsubscribeError: When unsubscribed from a channel
        """
        self.logger.info("Unsubscribing from %s", channel)
        if (subscription := self._channels.get(channel)) is not None:
            await self.publish(channel, STOPWORD)
            await subscription.pubsub.unsubscribe(channel)
            subscription.task.cancel()
            await subscription.task
            del self._channels[channel]
        else:
            self.logger.warning("Unknown channel %s", channel)
            raise UnsubscribeError(f"Unknown channel {channel}")

    async def publish(self, channel: str, message: Any) -> None:
        """Using the configured redis connection pool a publish a message to the provided channel

        :param channel: Channel name to use to publish the message to
        :type channel: str
        :param message: Data that is to be published to the channel
        :type message: Any
        """
        self.logger.debug("Publishing message %s to channel %s", message, channel)
        await self._redis.publish(channel, message)

    async def next_event(self) -> Union[Event, None]:
        """Return the next event from the queue that was read from all channels"""
        self.logger.debug("Getting next event from redis...")
        return await self._listen_queue.get()
